﻿if ( !Array.prototype.forEach ) {
	Array.prototype.forEach = function( callback, thisArg ) {
		var T, k;
		if ( this == null ) {
			throw new TypeError( " this is null or not defined" );
		}
		// 1. Let O be the result of calling ToObject passing the |this| value as the argument.
		var O = Object(this);
		// 2. Let lenValue be the result of calling the Get internal method of O with the argument "length".
		// 3. Let len be ToUint32(lenValue).
		var len = O.length >>> 0; // Hack to convert O.length to a UInt32
		// 4. If IsCallable(callback) is false, throw a TypeError exception.
		// See: http://es5.github.com/#x9.11
		if ( {}.toString.call(callback) != "[object Function]" ) {
			throw new TypeError( callback + " is not a function" );
		}
		// 5. If thisArg was supplied, let T be thisArg; else let T be undefined.
		if ( thisArg ) {
			T = thisArg;
		}
		// 6. Let k be 0
		k = 0;

		// 7. Repeat, while k < len
		while( k < len ) {
			var kValue;
			// a. Let Pk be ToString(k).
			//   This is implicit for LHS operands of the in operator
			// b. Let kPresent be the result of calling the HasProperty internal method of O with argument Pk.
			//   This step can be combined with c
			// c. If kPresent is true, then
			if ( k in O ) {
				// i. Let kValue be the result of calling the Get internal method of O with argument Pk.
				kValue = O[ k ];
				// ii. Call the Call internal method of callback with T as the this value and
				// argument list containing kValue, k, and O.
				callback.call( T, kValue, k, O );
			}
			// d. Increase k by 1.
			k++;
		}
		// 8. return undefined
	};
}

if (Object.assign === undefined) {
	Object.prototype.assign = function(target, source) {
//		for (var source of sources) {
			var keys = Object.getOwnPropertyNames(source);
			for (var p in keys) {
				var key = keys[p];
				var desc = Object.getOwnPropertyDescriptor(source, key);
				if (desc && desc.enumerable) {
					target[key] = source[key];
				}
			}
//		}
	};
}


function UI_Rect(x, y, w, h) {
	var rect = Object.create(UI_Rect);
	rect.x = +x || 0;
	rect.y = +y || 0;
	rect.w = +w || 0;
	rect.h = +h || 0;
	return rect;
}

UI_Rect.isinside = function(x, y) {
	x = +x || 0;
	y = +y || 0;
	if (this.x <= x && x < this.x + this.w && this.y <= y && y < this.y + this.h) {
		return true;
	}
	else {
		return false;
	}
};

function UI_Style(param) {
	var style = Object.create(UI_Style);
	style.font	=					(param && param.font)					||	"Century, sans serif";
	style.fontsize =			(param && param.fontsize)			|| 12;
	style.textStyle =			(param && param.textStyle)		|| "black";
	style.strokeStyle =		(param && param.strokeStyle)	|| "black";
	style.lineWidth =			(param && param.lineWidth)		|| 0;
	style.fillStyle =			(param && param.fillStyle)		|| "transparent";
	style.align =					(param && param.align)				|| "left";
	style.valign =				(param && param.valign)				|| "middle";
	style.px =						(param && param.px)						|| 0;
	style.py =						(param && param.py)						|| 0;
	return style;
}

function UI_Drawdata(type, param) {
	var drawdata =						Object.create(UI_Drawdata);
	drawdata.type =						type;
	drawdata.id =							(param && param.id !== undefined) ? param.id : drawdata.type;
	drawdata.value =					(param && param.value !== undefined) ? param.value : null;
	drawdata.style =					UI_Style(param && param.style);
	drawdata.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : drawdata.style;
	drawdata.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : drawdata.style;
	drawdata.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : drawdata.offstyle;
	drawdata.x =							(param && param.x) || 0;
	drawdata.y = 							(param && param.y) || 0;
	drawdata.width =					(param && param.width) || 50;
	drawdata.height	=					(param && param.height) || 50;
	drawdata.position =				(param && param.position) || "static";
	drawdata.transform =			(param && param.transform) || null;
	drawdata.param =					(param && param.param) || null;
	drawdata.rect =						UI_Rect();
	drawdata.mouseover = false;
	drawdata.on	= true;
	return drawdata;
}

UI_Drawdata.prototype.gettype = function() {
	return this.type;
},
UI_Drawdata.prototype.update = function() {
	this.rect =		UI_Rect(this.rect.x, this.rect.y, this.width, this.height);
}
UI_Drawdata.prototype.getstyle = function() {
	return (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
}
UI_Drawdata.prototype.drawbackground = 	function(ctx) {
	var style = this.getstyle();
	ctx.beginPath();
	ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
	ctx.fillStyle = style.fillStyle;
	ctx.fill();
	if (style.lineWidth > 0) {
		ctx.beginPath();
		ctx.lineWidth		= style.lineWidth;
		ctx.strokeStyle	= style.strokeStyle;
		ctx.rect(this.rect.x + style.lineWidth / 2, this.rect.y + style.lineWidth / 2, this.rect.w - style.lineWidth, this.rect.h - style.lineWidth);
		ctx.strokeStyle	= style.strokeStyle;
		ctx.stroke();
	}
}

function UI_Drawdata_Text(param) {
	var ddata = Object.create(UI_Drawdata_Text.prototype);
	Object.assign(ddata, UI_Drawdata("text", param));
	ddata.text				= (param && param.text) || "";
	ddata.bindvar			= (param && param.bindvar) || null;
	return ddata;
}

UI_Drawdata_Text.prototype.draw =	function(ctx) {
	this.drawbackground(ctx);
	var style = this.getstyle();
	ctx.fillStyle			= style.textStyle;
	ctx.textBaseline	= "bottom";
	ctx.font					= style.fontsize + "px " + style.font;
	var text = this.text = (this.bindvar) ? eval(this.bindvar) : this.text;
	var twidth = ctx.measureText(text).width;
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.fontsize + style.py;
	if (style.align === "center") {
		x = this.rect.x + (this.width - twidth) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w - twidth - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (style.fontsize + this.rect.h) / 2;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - style.lineWidth - style.py;
	}
	x |= 0;
	y |= 0;
	ctx.fillText(text, x, y);
}

Object.setPrototypeOf(UI_Drawdata_Text.prototype, UI_Drawdata.prototype);

function UI_Drawdata_Image(param) {
	var ddata = Object.create(UI_Drawdata_Image.prototype);
	Object.assign(ddata, UI_Drawdata("image", param));
	if (param && param.image) {
		ddata.image = param.image;
	}
	else {
		ddata.image	=				new Image();
		if (param && param.src) {
			ddata.image.src =		(param && param.src) || null;
			ddata.image.onload = function(param) {
				this.imagewidth =		this.imagewidth || this.image.width;
				this.imageheight =	this.imageheight || this.image.height;
			}.bind(ddata);
		}
	}
	ddata.image.ddata = ddata;
	ddata.imagewidth =	(param && param.imagewidth) || ddata.image.width;
	ddata.imageheight =	(param && param.imageheight) || ddata.image.height;

	ddata.imagesrcx		=	(param && param.imagesrcx) || 0;
	ddata.imagesrcy		=	(param && param.imagesrcy) || 0;
	ddata.imagesrcw		=	(param && param.imagesrcw) || 0;
	ddata.imagesrch		=	(param && param.imagesrch) || 0;
	return ddata;
}

UI_Drawdata_Image.prototype.draw = function(ctx) {
	this.drawbackground(ctx);
	var style = this.getstyle();
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.py;
	if (style.align === "center") {
		x = this.rect.x + (this.rect.w - this.imagewidth) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w -  this.imagewidth - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (this.rect.h - this.imageheight) / 2;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - this.imageheight - style.lineWidth - style.py;
	}
	if (this.transform) {
		ctx.save();
		if (this.transform === "reversex") {
			ctx.scale(-1, 1);
			ctx.translate(-x * 2 - this.imagewidth, 0);
		}
	}
	x |= 0;
	y |= 0;
	if (this.imagesrcw) {
		ctx.drawImage(this.image, this.imagesrcx, this.imagesrcy, this.imagesrcw, this.imagesrch, x, y, this.imagewidth, this.imageheight);
	}
	else {
		ctx.drawImage(this.image, x, y, this.imagewidth, this.imageheight);
	}
	if (this.transform) {
		ctx.restore();
	}
}

Object.setPrototypeOf(UI_Drawdata_Image.prototype, UI_Drawdata.prototype);

function UI_Drawdata_PlusIntegerImage(param) {
	var ddata = Object.create(UI_Drawdata_PlusIntegerImage.prototype);
	Object.assign(ddata, UI_Drawdata("plusintegerimage", param));
	if (param && param.image) {
		ddata.image = param.image;
	}
	else {
		ddata.image	=				[];
		if (param && param.src) {
			let index = param.src.lastIndexOf(".");
			let src = [ param.src.substr(0, index), param.src.substr(index + 1) ];
			for (let i = 0 ; i <= 9 ; i++) {
				ddata.image[i] = new Image();
				ddata.image[i].src =	src[0] + i + "." + src[1];
			}
		}
		ddata.image[0].onload = function() {
			this.imagewidth =		this.imagewidth || this.image[0].width;
			this.imageheight =	this.imageheight || this.image[0].height;
		}.bind(ddata);
	}
	ddata.text = (param && param.text) || 0;
	ddata.image.ddata = ddata;
	ddata.imagewidth =	(param && param.imagewidth) || (ddata.image && ddata.image[0] && ddata.image[0].width) || 0;
	ddata.imageheight =	(param && param.imageheight) || (ddata.image && ddata.image[0] && ddata.image[0].height) || 0;
	ddata.imagesrcx		=	(param && param.imagesrcx) || 0;
	ddata.imagesrcy		=	(param && param.imagesrcy) || 0;
	ddata.imagesrcw		=	(param && param.imagesrcw) || 0;
	ddata.imagesrch		=	(param && param.imagesrch) || 0;
	return ddata;
}

UI_Drawdata_PlusIntegerImage.prototype.draw = function(ctx) {
	this.drawbackground(ctx);
	var style = this.getstyle();
	var x = this.rect.x + style.lineWidth + style.px;
	var y = this.rect.y + style.lineWidth + style.py;

	var num = Math.floor(+this.text);
	if (!isFinite(num) || num < 0) {
		num = 0;
	}
	this.text = num + '';
	let imagewidth = this.imagewidth * this.text.length;

	if (style.align === "center") {
		x = this.rect.x + (this.rect.w - imagewidth) / 2;
	}
	else if (style.align === "right") {
		x = this.rect.x + this.rect.w -  imagewidth - style.lineWidth - style.px;
	}
	if (style.valign === "middle") {
		y = this.rect.y + (this.rect.h - this.imageheight) / 2;
	}
	else if (style.valign === "bottom") {
		y = this.rect.y + this.rect.h - this.imageheight - style.lineWidth - style.py;
	}
	if (this.transform) {
		ctx.save();
		if (this.transform === "reversex") {
			ctx.scale(-1, 1);
			ctx.translate(-x * 2 - imagewidth, 0);
		}
	}
	x |= 0;
	y |= 0;
	for (let i = 0 ; i < this.text.length ; i++) {
		ctx.drawImage(this.image[this.text.substr(i, 1)], x + this.imagewidth * i, y, this.imagewidth, this.imageheight);
	}
	if (this.transform) {
		ctx.restore();
	}
}

Object.setPrototypeOf(UI_Drawdata_PlusIntegerImage.prototype, UI_Drawdata.prototype);


UI2.eventlist = [
	"mousedown",
	"mouseup",
	"mousemove",
	"mouseout",
	"wheel",
];

function UI2(type, param) {
	var ui = Object.create(UI2.prototype);
	ui.type =						type;
	ui.id =							(param && param.id !== undefined) ? param.id : ui.type;
	ui.value =					(param && param.value) || null;
	ui.x =							(param && param.x) || 0;
	ui.y = 							(param && param.y) || 0;
	ui.px =							(param && param.px) || 0;
	ui.py = 						(param && param.py) || 0;
	ui.param = 					(param && param.param !== undefined) ? param.param : null;
	ui.position = 			(param && param.position) || "static";
	ui.width	=					(param && param.width) || 0;
	ui.height	=					(param && param.height) || 0;
	ui.autosize =				(param && param.autosize) || true;
	ui.style =					(param && param.style) ? UI_Style(param.style) : null;
	ui.moverstyle =			(param && param.moverstyle) ? UI_Style(param.moverstyle) : ui.style;
	ui.offstyle =				(param && param.offstyle) ? UI_Style(param.offstyle) : ui.style;
	ui.moveroffstyle =	(param && param.moveroffstyle) ? UI_Style(param.moveroffstyle) : ui.offstyle;
	ui.arg =						(param && param.arg) || null;
	ui.canvas =					(param && param.canvas) || null;
	ui.ctx =						(param && param.ctx) || null;
	ui.data	=						(param && param.data) || [];
	ui.display =				(param && param.display) || true;
	ui.enablewheel =		(param && param.enablewheel) || false;
	ui.rect =						UI_Rect();
	if (param && param.preprocess) {
		ui.preprocess = param.preprocess;
	}
	if (param && param.bindvar) {
		ui.bindvar = param.bindvar;
	}
	if (param && param.onchange) {
		ui.onchange = param.onchange;
	}
	UI2.eventlist.forEach(function(data, i, a) {
		if (param && param["on" + data]) {
			ui["on" + data] = param["on" + data];
		}
	});


	if (ui.data.forEach === undefined) {
		ui.data = [ ui.data ];
	}
	ui.mouseover =	false;
	return ui;
}



UI2.prototype.getstyle = function() {
	return (this.on) ? ((this.mouseover) ? this.moverstyle : this.style) : ((this.mouseover) ? this.moveroffstyle : this.offstyle);
}
UI2.prototype.update = function() {
	this.rect =		UI_Rect(this.x, this.y, this.width, this.height);
	this.data.forEach(function(data, i, a) {
		data.rect.x =		this.x;
		data.rect.y =		this.y;
		data.arg =			this.arg;
		data.base =			this.base;
		data.update();
	}, this);
}
UI2.prototype.drawcommon = function() {
	if (!this.display) {
		return;
	}
	var ctx = this.ctx;
	var style = this.getstyle();
	if (style !== null) {
		if (style.fillStyle !== "transparent") {
			ctx.beginPath();
			ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
			ctx.fillStyle = style.fillStyle;
			ctx.fill();
		}
	}
	this.draw();
	if (style !== null) {
		if (style.lineWidth > 0) {
			ctx.beginPath();
			ctx.lineWidth = style.lineWidth;
			ctx.strokeStyle = style.strokeStyle;
			ctx.rect(this.rect.x + style.lineWidth / 2, this.rect.y + style.lineWidth / 2, this.rect.w - style.lineWidth, this.rect.h - style.lineWidth);
			ctx.strokeStyle = style.strokeStyle;
			ctx.stroke();
		}
	}
}
UI2.prototype.updatebindvar = function() {
}
UI2.prototype.preprocessorig = function() {
}
UI2.prototype.preprocess = function() {
}
UI2.prototype.checknewvalue = function() {
	return true;
}
UI2.prototype.onchange =  function() {
}
UI2.prototype.draw =  function() {
	if (this.data.length > 0) {
		this.data[0].draw(this.ctx);
	}
}
UI2.prototype.paint = function(preprocessflag) {
	if (this.display === false) {
		return;
	}
	if (preprocessflag !== false) {
		this.base.mousemovepreprocess(null);
	}
	let ctx = this.ctx;
	ctx.fillStyle = "rgb(240, 240, 240)";
	ctx.beginPath();
	ctx.rect(this.rect.x, this.rect.y, this.rect.w, this.rect.h);
	ctx.fill();
	this.drawcommon();
}
UI2.prototype.getdatafromid = function(id) {
	for (let i = 0 ; i < this.data.length ; i++) {
		if (this.data[i].id === id) {
			return this.data[i];
		}
	}
	return null;
}

function UI_Container(param) {
	var container = Object.create(UI_Container.prototype);
	Object.assign(container, UI2("container", param));
	container.space =					(param && param.space) || 0;
	container.lineheight =		(param && param.lineheight) || 0;
	container.update();
	return container;
}

UI_Container.prototype.update =	function() {
	if (this.type === "base") {
		this.rect.x = this.x;
		this.rect.y = this.y;
	}
	var x = this.rect.x + this.px;
	var y = this.rect.y + this.py;
	var maxx = 0;
	var maxy = 0;

	this.data.forEach(function(data, i, a) {
		if (data.position === "absolute") {
			data.rect.x = this.rect.x + data.x;
			data.rect.y = this.rect.y + data.y;
		}
		else if (data.position === "relative") {
			x += data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
		else if (data.position === "newline") {
			x = this.rect.x + this.px + data.x;
			y += data.y + this.lineheight;
			data.rect.x = x;
			data.rect.y = y;
		}
		else {
			data.rect.x =		x;
			data.rect.y =		y;
		}
		if (maxx < data.rect.x + data.rect.w) {
			maxx = data.rect.x + data.rect.w;
		}
		if (maxy < data.rect.y + data.rect.h) {
			maxy = data.rect.y + data.rect.h;
		}
		data.ctx =		this.ctx;
		data.canvas =	this.canvas;
		data.arg =		this.arg;
		data.base	=		this.base;
		data.update();
		if (data.position !== "absolute") {
			x += data.rect.w + this.space;
		}
	}, this);
	if (this.autosize) {
		this.rect.w = maxx - this.rect.x;
		this.rect.h = maxy - this.rect.y;
	}
}
UI_Container.prototype.draw = function() {
	this.data.forEach(function(data, i, a) {
		data.drawcommon();
	});
}
UI_Container.prototype.preprocessorig = function() {
	var bindvar = this.bindvar;
	var bindvarvalue = bindvar && eval(bindvar);
	this.data.forEach(function(data, i, a) {
		if (bindvar) {
			data.display = bindvarvalue === data.value;
		}
	});
}
UI_Container.prototype.getdatafromid = function(id) {
	for (let i = 0 ; i < this.data.length ; i++) {
		if (this.data[i].id === id) {
			return this.data[i];
		}
		let retval = this.data[i].getdatafromid(id);
		if (retval) {
			return retval;
		}
	}
	return null;
}

Object.setPrototypeOf(UI_Container.prototype, UI2.prototype);

function UI_Base(param) {
	var container = Object.create(UI_Base.prototype);
	Object.assign(container, UI_Container(param));
	container.type = "base";
	container.base = container;
	if (container.canvas) {
		var canvas = container.canvas;
		canvas.ui = container;
		container.eventorig = {};
		UI2.eventlist.forEach(function(data, i, a) {
			canvas.addEventListener(data, container[data + "base"]);
		}, this);
	}
	container.update();
	return container;
}

UI2.eventlist.forEach(function(data, i, a) {
	UI_Base.prototype[data + "base"] = function(evt) {
		var brect = this.getBoundingClientRect();
		evt.mx = Math.floor(evt.clientX - brect.left);
		evt.my = Math.floor(evt.clientY - brect.top);
		this.ui.mouseover = this.ui.rect.isinside(evt.mx, evt.my);

		this.ui[data + "preprocess"](evt);
		this.ui[data](evt);
		this.ui["on" + data].apply(this.ui, [ evt ]);
		this.ui.paint(false);
	};

	UI2.prototype[data + "preprocess"] = function(evt) {
		this.preprocess();
		if (!this.display) {
			return;
		}
		this.preprocessorig();
		this.data.forEach(function(data2, i, a) {
			if (evt) {
				data2.mouseover = data2.rect.isinside(evt.mx, evt.my);
			}
			if (this.type === "container" || this.type === "base") {
				data2[data + "preprocess"](evt);
				if (data2.display && evt) {
					if (data === "wheel" || data2.mouseover) {
						data2[data](evt);
						data2["on" + data](evt);
					}
				}
			}
		}, this);
	};
	UI2.prototype[data] = function(evt) {
	};
	UI2.prototype["on" + data] = function(evt) {
	};
});

Object.setPrototypeOf(UI_Base.prototype, UI_Container.prototype);

function UI_Button(param) {
	var button = Object.create(UI_Button.prototype);
	Object.assign(button, UI2("button", param));
	if (!param || !param.moveroffstyle) {
		button.moveroffstyle = button.moverstyle;
	}
	button.update();
	return button;
}

UI_Button.prototype.update = function() {
	if (this.data.length > 0) {
		this.data[0].arg 		=	this.arg;
		this.data[0].rect.x =	this.rect.x + this.px;
		this.data[0].rect.y =	this.rect.y + this.py;
		this.data[0].update();
		if (this.autosize) {
			this.rect.w = this.px * 2 + this.data[0].width;
			this.rect.h = this.py * 2 + this.data[0].height;
		}
	}
}

Object.setPrototypeOf(UI_Button.prototype, UI2.prototype);

function UI_SelectorCommon(type, param) {
	var selector = Object.create(UI_SelectorCommon.prototype);
	Object.assign(selector, UI2(type, param));
	selector.selectnum =		(param && param.selectnum) || 0;
	selector.selectvalue = "none";
	return selector;
}
UI_SelectorCommon.prototype.preprocessorig = function() {
	if (this.arg && this.bindvar) {
		var selectvalue = eval(this.bindvar);
		if (this.selectvalue !== selectvalue) {
			var selectnum = -1;
			this.data.forEach(function(data, i, a) {
				if (data.value === selectvalue) {
					selectnum = i;
				}
			});
			if (selectnum >= 0) {
				this.selectvalue = selectvalue;
				this.selectnum = selectnum;
				this.onchange();
			}
			else {
				this.updatebindvar();
			}
		}
	}
}
UI_SelectorCommon.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.selectvalue;");
	}
}
UI_SelectorCommon.prototype.wheel = function(evt) {
	if (this.data.length > 0) {
		if (this.enablewheel === true || (this.enablewheel === "mouseover" && this.rect.isinside(evt.mx, evt.my))) {
			if (evt.deltaY > 0) {
				this.selectnum = (this.selectnum + 1) % this.data.length;
			}
			else if (evt.deltaY < 0) {
				this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
			}
			else {
				return;
			}
			this.selectvalue = this.data[this.selectnum].value;
			this.updatebindvar();
			this.onchange();
		}
	}
}

Object.setPrototypeOf(UI_SelectorCommon.prototype, UI2.prototype);

function UI_Selector(param) {
	var selector = Object.create(UI_Selector.prototype);
	Object.assign(selector, UI_SelectorCommon("selector", param));
	selector.space =				(param && param.space) || 0;
	selector.lineheight =		(param && param.lineheight) || 10;
	selector.update();
	return selector;
}

UI_Selector.prototype.update= function() {
	var x =		this.rect.x + this.px;
	var y =		this.rect.y + this.py;
	var maxx	= 0;
	var maxy	= 0;
	var selectnum = +this.selectnum;
	if (selectnum < 0 && selectnum >= this.data.length) {
		selectnum = this.selectnum = 0;
	}
	if (this.data.length > 0) {
		this.selectvalue = this.data[selectnum].value;
	}
	else {
		this.selectvalue = "";
	}
	this.data.forEach(function(data, i, a) {
		if (data.position === "absolute") {
			data.rect.x = this.rect.x + data.x;
			data.rect.y = this.rect.y + data.y;
		}
		else if (data.position === "relative") {
			x += data.x;
			y += data.y;
			data.rect.x = x;
			data.rect.y = y;
		}
		else if (data.position === "newline") {
			x = this.rect.x + this.px + data.x;
			y += data.y + this.lineheight;
			data.rect.x = x;
			data.rect.y = y;
		}
		else {
			data.rect.x =		x;
			data.rect.y =		y;
		}
		if (maxx < data.rect.x + data.rect.w) {
			maxx = data.rect.x + data.rect.w;
		}
		if (maxy < data.rect.y + data.rect.h) {
			maxy = data.rect.y + data.rect.h;
		}

		data.arg =	this.arg;
		data.base = this.base;
		data.update();
		data.on = (i == selectnum);
		if (data.position !== "absolute") {
			x += data.rect.w + this.space;
		}
	}, this);
	if (this.autosize) {
		this.rect.w = maxx - this.rect.x;
		this.rect.h = maxy - this.rect.y;
	}
}
UI_Selector.prototype.draw =function() {
	let selectnum = this.selectnum;
	this.data.forEach(function(data, i, a) {
		data.on = (selectnum == i);
		data.draw(this.ctx);
	}, this);
}
UI_Selector.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	var newselectnum = -1;
	this.data.forEach(function(data, i, a) {
		if (data.mouseover) {
			newselectnum = i;
		}
	});
	if (newselectnum != -1 && newselectnum != this.selectnum) {
		this.selectnum = newselectnum;
		this.selectvalue = this.data[this.selectnum].value;
		this.updatebindvar();
		this.onchange();
	}
}

Object.setPrototypeOf(UI_Selector.prototype, UI_SelectorCommon.prototype);

function UI_ClickSelector(param) {
	var selector = Object.create(UI_ClickSelector.prototype);
	Object.assign(selector, UI_SelectorCommon("clickselector", param));
	selector.selectnum =		(param && param.selectnum) || 0;
	selector.selectvalue = null;
	if (!param || !param.moveroffstyle) {
		selector.moveroffstyle = selector.moverstyle;
	}
	selector.update();
	return selector;
}

UI_ClickSelector.prototype.update = function() {
	var x =		this.rect.x + this.px;
	var y =		this.rect.y + this.py;
	var selectnum = +this.selectnum;
	if (selectnum < 0 && selectnum >= this.data.length) {
		selectnum = this.selectnum = 0;
	}
	if (this.data.length > 0) {
		this.selectvalue = this.data[selectnum].value;
	}
	else {
		this.selectvalue = null;
	}
	var maxw = 0;
	var maxh = 0;
	this.data.forEach(function(data, i, a) {
		data.rect.x =		x;
		data.rect.y =		y;
		data.arg =			this.arg;
		data.base =			this.base;
		if (maxw < data.rect.w) {
			maxw = data.rect.w;
		}
		if (maxh < data.rect.h) {
			maxh = data.rect.h;
		}
		data.update();
	});
	if (this.autosize) {
		this.rect.w = maxw + this.px * 2;
		this.rect.h = maxh + this.py * 2;
	}
}
UI_ClickSelector.prototype.draw = function(ui) {
	if (this.data.length > 0) {
		this.data[this.selectnum].draw(this.ctx);
	};
}
UI_ClickSelector.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	if (this.data.length > 0) {
		if (evt.buttons & 1) {
			this.selectnum = (this.selectnum + 1) % this.data.length;
		}
		else if (evt.buttons & 2) {
			this.selectnum = (this.selectnum + this.data.length - 1) % this.data.length;
		}
		this.selectvalue = this.data[this.selectnum].value;
		this.updatebindvar();
		this.onchange();
	}
}

Object.setPrototypeOf(UI_ClickSelector.prototype, UI_SelectorCommon.prototype);

function UI_TextBox(param) {
	var textbox = Object.create(UI_TextBox.prototype);
	Object.assign(textbox, UI2("textbox", param));
	if (!param || !param.moveroffstyle) {
		textbox.moveroffstyle = textbox.moverstyle;
	}
	textbox.value		= (param && param.value) || "";
	textbox.msg			= (param && param.msg) || "";
	textbox.update();
	return textbox;
}

UI_TextBox.prototype.update = function() {
	if (this.data.length > 0) {
		this.data[0].arg =		this.arg;
		this.data[0].base =		this.base;
		this.data[0].rect.x =	this.rect.x + this.px;
		this.data[0].rect.y =	this.rect.y + this.py;
		this.data[0].update();
		if (this.autosize) {
			this.rect.w = this.px * 2 + this.data[0].rect.w;
			this.rect.h = this.py * 2 + this.data[0].rect.h;
		}
	}
}
UI_TextBox.prototype.preprocessorig = function() {
	if (this.arg && this.bindvar) {
		this.value = eval(this.bindvar);
	}
	if (this.data.length > 0) {
			this.data[0].text = this.value;
	}
}
UI_TextBox.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.value;");
	}
}
UI_TextBox.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	let text = window.prompt(this.msg, this.value);
	if (text !== null) {
		this.newvalue = text;
		if (this.checknewvalue()) {
			this.value = text;
			this.updatebindvar();
			this.onchange();
		}
	}
}

Object.setPrototypeOf(UI_TextBox.prototype, UI2.prototype);

function UI_NumberBox(param) {
	var numbox = Object.create(UI_NumberBox.prototype);
	Object.assign(numbox, UI2("numberbox", param));
	if (!param || !param.moveroffstyle) {
		numbox.moveroffstyle = numbox.moverstyle;
	}
	numbox.value		= (param && param.value) || 0;
	numbox.msg			= (param && param.msg) || "";
	numbox.numtype	= (param && param.numtype) || "real";
	numbox.max			= (param && param.max) || Infinity;
	numbox.min			= (param && param.min) || -Infinity;
	numbox.wheelnum	= (param && param.wheelnum) || 1;
	numbox.update();
	return numbox;
}

UI_NumberBox.prototype.update = function() {
	if (this.data.length > 0) {
		this.data[0].arg =		this.arg;
		this.data[0].base =		this.base;
		this.data[0].rect.x =	this.rect.x + this.px;
		this.data[0].rect.y =	this.rect.y + this.py;
		this.data[0].update();
		if (this.autosize) {
			this.rect.w = this.px * 2 + this.data[0].rect.w;
			this.rect.h = this.py * 2 + this.data[0].rect.h;
		}
	}
}

UI_NumberBox.prototype.changevalue = function(value) {
	this.newvalue = value;
	if (this.numtype === "integer") {
		this.newvalue	|= 0;
		this.min			|= 0;
		this.max			|= 0;
	}
	else {
		this.newvalue	= +value;
		this.min			= +this.min;
		this.max			= +this.max;
	}
	if (this.min > this.max) {
		this.min = this.max;
	}
	if (this.newvalue < this.min) {
		this.newvalue = this.min;
	}
	else if (this.newvalue > this.max) {
		this.newvalue = this.max;
	}
	if (this.newvalue === this.value) {
		return;
	}
	if (this.checknewvalue()) {
		this.value = this.newvalue;
		this.updatebindvar();
		if (this.data.length > 0) {
				this.data[0].text = this.value;
		}
		this.onchange();
	}
}

UI_NumberBox.prototype.preprocessorig = function() {
	let newvalue = this.value;
	if (this.arg && this.bindvar) {
		newvalue = eval(this.bindvar);
	}
	if (newvalue !== this.value) {
		this.changevalue(newvalue);
	}
}
UI_NumberBox.prototype.updatebindvar = function() {
	if (this.arg && this.bindvar) {
		eval(this.bindvar + " = this.value;");
	}
}
UI_NumberBox.prototype.mousedown = function(evt) {
	if (!(evt.buttons & 1)) {
		return;
	}
	let num = window.prompt(this.msg, this.value);
	if (num !== null) {
		if (this.numtype === "integer") {
			num = num | 0;
		}
		else {
			num = +num;
		}
		this.changevalue(num);
	}
}
UI_NumberBox.prototype.wheel = function(evt) {
	if (this.enablewheel === true || (this.enablewheel === "mouseover" && this.rect.isinside(evt.mx, evt.my))) {
		if (evt.deltaY > 0) {
			this.changevalue(this.value + this.wheelnum);
		}
		else if (evt.deltaY < 0) {
			this.changevalue(this.value - this.wheelnum);
		}
		else {
			return;
		}
	}
}

Object.setPrototypeOf(UI_NumberBox.prototype, UI2.prototype);

